package ru.dbotthepony.mc.otm.client

import com.mojang.blaze3d.platform.InputConstants
import com.mojang.blaze3d.systems.RenderSystem
import net.minecraft.client.KeyMapping
import net.minecraftforge.client.event.RegisterKeyMappingsEvent
import net.minecraftforge.client.event.RenderGuiEvent
import net.minecraftforge.client.settings.KeyConflictContext
import org.lwjgl.glfw.GLFW.GLFW_CURSOR_DISABLED
import org.lwjgl.glfw.GLFW.GLFW_CURSOR_NORMAL
import ru.dbotthepony.mc.otm.OverdriveThatMatters
import ru.dbotthepony.mc.otm.android.AndroidFeatureType
import ru.dbotthepony.mc.otm.android.AndroidSwitchableFeature
import ru.dbotthepony.mc.otm.capability.matteryPlayer
import ru.dbotthepony.mc.otm.client.render.drawArc
import ru.dbotthepony.mc.otm.core.linearInterpolation
import ru.dbotthepony.mc.otm.network.MatteryPlayerNetworkChannel
import ru.dbotthepony.mc.otm.network.SwitchAndroidFeaturePacket
import java.util.stream.Collectors
import kotlin.math.PI
import kotlin.math.acos
import kotlin.math.cos
import kotlin.math.pow
import kotlin.math.sin

// TODO: forge's KeyMapping patch (ALL and MAP maps) is not thread safe... should we be worried?
object AndroidMenuKeyMapping : KeyMapping("key.otm.android_menu", KeyConflictContext.IN_GAME, InputConstants.Type.KEYSYM.getOrCreate(InputConstants.KEY_F), OverdriveThatMatters.MOD_ID) {
	private var grabbedInput = false
	private var selectedFeature: AndroidSwitchableFeature? = null
	private var lastSelectedFeature: AndroidSwitchableFeature? = null

	private var lastRender: Long = System.nanoTime()
	private var lastSelectedDegree: Double? = null
	private var lastSelectProgress = 0f

	override fun setDown(isDown: Boolean) {
		val old = this.isDown
		super.setDown(isDown)

		if (old != isDown) {
			if (
				minecraft.screen == null &&
				isDown &&
				minecraft.player?.matteryPlayer?.isAndroid == true &&
				minecraft.player?.matteryPlayer?.features?.anyMatch { it is AndroidSwitchableFeature && it.allowToSwitchByPlayer } == true
			) {
				grabbedInput = true
				minecraft.mouseHandler.releaseMouse()
			} else if (!isDown && grabbedInput) {
				grabbedInput = false

				if (minecraft.screen == null) {
					minecraft.mouseHandler.grabMouse()

					val selectedFeature = selectedFeature

					if (selectedFeature != null) {
						MatteryPlayerNetworkChannel.sendToServer(SwitchAndroidFeaturePacket(selectedFeature.type, !selectedFeature.isActive))
					}
				}
			}

			selectedFeature = null
			lastSelectedFeature = null
			lastSelectedDegree = null
			lastSelectProgress = 0f
		}
	}

	fun onRenderGuiEvent(event: RenderGuiEvent.Post) {
		if (!grabbedInput) {
			lastRender = System.nanoTime()
			return
		}

		val features = minecraft.player?.matteryPlayer?.features?.filter { it is AndroidSwitchableFeature && it.allowToSwitchByPlayer }?.collect(Collectors.toList()) as MutableList<AndroidSwitchableFeature>? ?: return

		if (features.isEmpty()) {
			lastRender = System.nanoTime()
			return
		}

		val delta = (System.nanoTime() - lastRender) / 1_000_000.0
		lastRender = System.nanoTime()

		RenderSystem.setShaderColor(0f, 0f, 0f, 0.75f)

		val size = minecraft.window.guiScaledHeight.coerceAtMost(minecraft.window.guiScaledWidth).toFloat() * 0.35f

		drawArc(
			event.poseStack,
			minecraft.window.guiScaledWidth / 2f,
			minecraft.window.guiScaledHeight / 2f,
			size,
			size * 0.3f
		)

		RenderSystem.setShaderColor(85 / 255f, 197 / 255f, 255 / 255f, 0.5f)

		val degreePerSlice = PI * 2.0 / features.size

		var (mouseX, mouseY) = mousePos

		mouseX -= minecraft.window.width / 2.0
		mouseY -= minecraft.window.height / 2.0

		val length = (mouseX * mouseX + mouseY * mouseY).pow(0.5)

		if ((length / minecraft.window.guiScale) in (size * 0.3f) .. size) {
			var deg = acos(mouseX / length)

			// opengl
			// lower values are at bottom of screen
			if (mouseY > 0.0) {
				deg = -deg
			}

			if (deg < 0.0) {
				deg += PI * 2
			}

			val index = (deg / degreePerSlice).toInt()
			selectedFeature = features[index]
			lastSelectedFeature = features[index]

			lastSelectProgress = (lastSelectProgress + delta * 0.01).coerceAtMost(1.0).toFloat()

			RenderSystem.setShaderColor(85 / 255f, 197 / 255f, 255 / 255f, 0.3f * lastSelectProgress)

			if (lastSelectedDegree == null) {
				this.lastSelectedDegree = index * degreePerSlice
			} else {
				this.lastSelectedDegree = linearInterpolation(delta, lastSelectedDegree!!, index * degreePerSlice)
			}
		} else {
			selectedFeature = null
			lastSelectProgress = (lastSelectProgress - delta * 0.01).coerceAtLeast(0.0).toFloat()

			if (lastSelectProgress == 0f) {
				lastSelectedDegree = null
				lastSelectedFeature = null
			}
		}

		if (lastSelectedDegree != null) {
			RenderSystem.setShaderColor(85 / 255f, 197 / 255f, 255 / 255f, 0.3f * lastSelectProgress)

			drawArc(
				event.poseStack,
				minecraft.window.guiScaledWidth / 2f,
				minecraft.window.guiScaledHeight / 2f,
				linearInterpolation(lastSelectProgress, size, size * 1.2f),
				linearInterpolation(lastSelectProgress, size * 0.3f, size * 0.4f),

				startDegree = lastSelectedDegree!!,
				endDegree = lastSelectedDegree!! + degreePerSlice
			)
		}

		RenderSystem.setShaderColor(1f, 1f, 1f, 1f)

		val iconSize = size * 0.25f
		val iconSizeSelected = linearInterpolation(lastSelectProgress, size * 0.25f, size * 0.3f)

		event.poseStack.pushPose()
		event.poseStack.translate(minecraft.window.guiScaledWidth.toDouble() / 2f, minecraft.window.guiScaledHeight.toDouble() / 2f, 0.0)

		if (features.size == 1) {
			val feature = features.first()

			feature.renderIcon(event.poseStack, -iconSizeSelected / 2f, -size * linearInterpolation(lastSelectProgress, 0.7f, 0.8f) - iconSizeSelected / 2f, iconSizeSelected, iconSizeSelected)
		} else {
			for ((index, feature) in features.withIndex()) {
				val sin = sin((index + 0.5) * degreePerSlice).toFloat()
				val cos = cos((index + 0.5) * degreePerSlice).toFloat()

				if (feature == lastSelectedFeature) {
					val selectedShift = linearInterpolation(lastSelectProgress, size * 0.7f, size * 0.8f)
					feature.renderIcon(event.poseStack, -iconSizeSelected / 2f + selectedShift * cos, -selectedShift * sin - iconSizeSelected / 2f, iconSizeSelected, iconSizeSelected)
				} else {
					feature.renderIcon(event.poseStack, -iconSize / 2f + size * 0.7f * cos, -size * 0.7f * sin - iconSize / 2f, iconSize, iconSize)
				}
			}
		}

		event.poseStack.popPose()
	}

	fun register(event: RegisterKeyMappingsEvent) {
		event.register(this)
	}
}
